/*
 * Copyright (c) 1991-2003 Rational Software Corporation. All rights
 * reserved.
 * 
 * Rational, the Rational logo, ObjecTime (and logo) are registered
 * trademarks of Rational Software Corporation.
 * 
 * The placing of the copyright notices on Rational source code is not
 * intended to mean that they are published works, but are included in the
 * event of inadvertent publication. The source code of Rational contains
 * trade secrets and other valuable confidential information of Rational
 * Software Corporation. Rational Software Corporation and its licensors
 * retain ownership to Rational Products in both source and object code
 * format, and related documentation and all rights therein including
 * copyrights, patents and trade secrets.
 * 
 * 
 * U.S. Government Rights
 * 
 * Use, duplication, or disclosure by the U.S. Government is subject to
 * restrictions set forth in the applicable Rational License Agreement
 * and in DFARS 227.7202-1(a) and 227.7202-3(a) (1995), DFARS
 * 252.227-7013(c)(1)(ii)(Oct 1988), FAR 12.212(a) 1995, FAR 52.227-19,
 * or FAR 52.227-14, as applicable.
 * 
 * Unless otherwise agreed to in writing, use of Rational products is
 * governed by the Rational Product License Agreement provided with
 * the product.
 */

#include <RTPriv/LayerCon.h>
#include <RTPriv/ResrcMgr.h>
#include <RTPriv/RTstdio.h>
#include <RTPubl/CapsRole.h>
#include <RTPubl/Capsule.h>
#include <RTPubl/Main.h>
#include <RTPubl/MemUtil.h>
#include <RTPubl/Port.h>
#include <RTPubl/ProtDesc.h>
#include <RTPubl/SystDesc.h>
#include <RTPubl/ThrdDesc.h>
#include <RTPubl/Version.h>

#if OTRTSDEBUG != DEBUG_NONE
#include <RTPriv/Debugger.h>
#endif

#if USE_THREADS
#include <RTPriv/Mutex.h>
#include <RTPriv/RTthread.h>
#include <RTPubl/PeerCont.h>

extern void RTPeerController_mainLoop_init( RTPeerController * );
#else
#include <RTPubl/SoleCont.h>

int RTMain_mainLine_finish( void );
#endif /* USE_THREADS */

/*
**  config_info tokens, (*) means the same token exists in the C++ TargetRTS.
**  -Cl    - ! RTS_CLEANUP_MECHANISM
**  -Df    - ! MESSAGE_DEFERRAL
**  -D     - ! OBJECT_DECODE (*)
**  -E     - ! OBJECT_ENCODE (*)
**  G0/1   - OTRTSDEBUG (*)
**  -Il    - ! INTERNAL_LAYER_SERVICE
**  -L     - ! LOG_MESSAGE (*)
**  +/-Me  - RTS_MEMORY_POLICY
**  -Na    - ! RTS_NAMES
**  -O     - ! OBSERVABLE (*)
**  +P     - PURIFY (*)
**  -Pr    - ! MULTIPLE_PRIORITIES
**  -Ti    - ! TIMING_SERVICE
**  -t     - ! USE_THREADS
**  -V     - ! SEND_BY_VALUE
**
**  config_info tokens used in the C++ TargetRTS for other purposes:
**  +C     - RTS_COUNT
**  F0/1   - RTFRAME_CHECKING
**  -Fs    - RTFRAME_THREAD_SAFE
**  -I     - HAVE_INET
**  +/-Tp  - RTTYPECHECK_PROTOCOL
**  +/-Tr  - RTTYPECHECK_RECEIVE
**  +/-Ts  - RTTYPECHECK_SEND
**  +b     - RTUseBitFields
**  +c     - RTS_COMPATIBLE
**  +d     - DEFER_IN_ACTOR (used for MESSAGE_DEFERRAL here)
**  +e     - EXTERNAL_LAYER
**  -i     - ! RTS_INLINES
**  -m     - ! RTMESSAGE_PAYLOAD_SIZE
**  -p     - ! INTEGER_POSTFIX
**  +s     - SUSPEND
**  -r     - ! RTREAL_INCLUDED
**  +/-Z   - RTStateId_MaxSize
*/

#if STDIO_ENABLED
static const char config_info[] =
{
#if ! RTS_CLEANUP_MECHANISM
    ',', '-', 'C', 'l',
#endif
#if ! MESSAGE_DEFERRAL
    ',', '-', 'D', 'f',
#endif
#if ! OBJECT_DECODE
    ',', '-', 'D',
#endif
#if ! OBJECT_ENCODE
    ',', '-', 'E',
#endif
#if   OTRTSDEBUG == DEBUG_NONE
    ',', 'G', '0',
#endif
#if ! INTERNAL_LAYER_SERVICE
    ',', '-', 'I', 'l',
#endif
#if ! LOG_MESSAGE
    ',', '-', 'L',
#endif
#if   RTS_MEMORY_POLICY == RTS_NEVER_ALLOCATE
    ',', '-', 'M', 'e',
#elif RTS_MEMORY_POLICY == RTS_CAN_ALLOCATE
    ',', '+', 'M', 'e',
#endif
#if ! RTS_NAMES
    ',', '-', 'N', 'a',
#endif
#if ! OBSERVABLE
    ',', '-', 'O',
#endif
#if PURIFY
    ',', '+', 'P',
#endif
#if ! MULTIPLE_PRIORITIES
    ',', '-', 'P', 'r',
#endif
#if ! TIMING_SERVICE
    ',', '-', 'T', 'i',
#endif
#if ! USE_THREADS
    ',', '-', 't',
#endif
#if ! SEND_BY_VALUE
    ',', '-', 'V',
#endif
    ')', '\0'
};

static void showBanner( void )
{
    RTStdio_putString( "Rational Rose RealTime C Target Run Time System\n" );
    RTStdio_putString( RT_VERSION_STRING );

    if( config_info[ 1 ] != '\0' )              /* more than ")"? */
    {
	RTStdio_putString( " (" );
	RTStdio_putString( &config_info[ 1 ] ); /* skip leading comma */
    }

    RTStdio_putString( "\nCopyright (c) 1993-" );
    RTStdio_putString( RT_VERSION_YEAR );
    RTStdio_putString( " Rational Software\n" );
}
#endif

int RTMain_mainLine_init( void )
{
    RTMain                   * this = RTMain_soleInstance();
    const RTSystemDescriptor * s    = RTMain_system();
    RTController             * mainController;
#if OTRTSDEBUG != DEBUG_NONE && USE_THREADS
    RTController             * debugController;
#endif
    const int                  numberOfThreads = RTMain_numberOfThreads();
#if OTRTSDEBUG != DEBUG_NONE || USE_THREADS
    int                        i;
#endif

    this->state            = RTMain_initializing;
#if USE_THREADS
    this->startupCountdown = numberOfThreads;
    this->mutex            = new_RTMutex();
#endif

    RTMain_startup();

#if STDIO_ENABLED
    showBanner();
#endif

#if INTERNAL_LAYER_SERVICE
    RTLayerConnector__init( MAX_NUM_SPPS );
#endif

    RTMain_installHandlers();

#if OBJECT_DECODE
    RTObject_class__startup();
    {
	int                            numTypes = s->numTypes;
	const RTObject_class * const * type     = s->types;

	for( ; --numTypes >= 0; ++type )
	    RTObject_class_install( *type );
    }
#endif

    this->controllers = (RTController **)
			RTMemoryUtil_new( numberOfThreads *
					  (int)sizeof( RTController * ) );

    RTResourceMgr_init();

#if USE_THREADS && STDIO_ENABLED
    RTStdio_startup();
#endif

    this->controllers[ 0 ] =
    mainController         =
#if USE_THREADS
			     new_RTPeerController
#else
			     new_RTSoleController
#endif
				 (
#if RTS_NAMES
				     "main",
#endif
				     (RTMessageIndex)DEFAULT_FREE_MSGQ_SIZE
				 );

#if USE_THREADS
    for( i = 1; i < numberOfThreads; ++i )
    {
	this->controllers[ i ] = new_RTPeerController(
#if RTS_NAMES
				  s->constThreads[ i ].name,
#endif
				  s->constThreads[ i ].numInternalMsgs );
    }
#endif /* USE_THREADS */

    RTCapsule__initTop();

#if OTRTSDEBUG != DEBUG_NONE
    RTDebugger_startup();

    for( i = 0; i < numberOfThreads; ++i )
	RTController_setDebugger( this->controllers[ i ] );

#if USE_THREADS
    this->startupCountdown += 1;
    debugController = RTDebugger_soleInstance()->controller;
    this->debugThread = new_RTThread( debugController,
				      DEFAULT_DEBUG_PRIORITY,
				      DEBUGGER_STACK_SIZE );
#else
    /* If debugger enabled and single-threaded, then initialize */
    RTDebugger_initialize();
#endif /* USE_THREADS */
#endif /* OTRTSDEBUG != DEBUG_NONE */

#if USE_THREADS
    this->threads = RTMemoryUtil_new( numberOfThreads *
				      (int)sizeof( RTThread * ) );

    for( i = 1; i < numberOfThreads; ++i )
    {
	this->threads[ i ] = new_RTThread( this->controllers[ i ],
					   s->constThreads[ i ].Priority,
					   s->constThreads[ i ].StackSize );
    }
#endif /* USE_THREADS */

#if USE_THREADS
    RTPeerController_mainLoop_init( mainController );
    RTPeerController_mainLoop( mainController );
#else
    (*mainController->mainLoop)( mainController );
#endif /* USE_THREADS */

#if USE_THREADS
    return 0;
#else
    return RTMain_mainLine_finish();
#endif
}

int RTMain_mainLine_finish( void )
{
    RTMain                   * this = RTMain_soleInstance();
    RTController             * mainController = this->controllers[ 0 ];
#if OTRTSDEBUG != DEBUG_NONE && USE_THREADS
    RTController             * debugController;
#endif
    const int                  numberOfThreads = RTMain_numberOfThreads();
#if OTRTSDEBUG != DEBUG_NONE || USE_THREADS
    int                        i;
#endif

/*    (*mainController->mainLoop)( mainController ); */
    this->controllers[ 0 ] = (RTController *)0;
    this->state            = RTMain_shuttingDown;

#if RTS_CLEANUP_MECHANISM
#if USE_THREADS
    for( i = 1; i < numberOfThreads; ++i )
    {
	delete_RTThread( this->threads[ i ], mainController );
	delete_RTPeerController( this->controllers[ i ] );
    }
#endif /* USE_THREADS */

#if OTRTSDEBUG != DEBUG_NONE
    RTDebugger_shutdown();
#if USE_THREADS
    delete_RTThread( this->debugThread, mainController );
    delete_RTPeerController( debugController );
#endif /* USE_THREADS */
#endif /* OTRTSDEBUG != DEBUG_NONE */

#if USE_THREADS
    delete_RTPeerController( mainController );
#else
    delete_RTSoleController( mainController );
#endif /* USE_THREADS */

    RTResourceMgr_shutdown();
    RTMain_shutdown();
#if USE_THREADS && STDIO_ENABLED
    RTStdio_shutdown();
#endif

#else /* ! RTS_CLEANUP_MECHANISM */
    /* we still need to kill all the jobs */
#if USE_THREADS
    for( i = 1; i < numberOfThreads; ++i )
	RTController_kill( this->threads[ i ]->_job, mainController );
#endif
#endif

    return 0;
}
